#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


"""
R3 Process Single Manifest helper.
"""

import logging
from typing import List, Tuple

from libs.constants import FIRST_STORED_RECORD_INDEX
from libs.context import Context
from libs.exceptions import ProcessRecordError
from libs.manifest_analyzer import ManifestAnalyzer, EntityNode
from libs.refresh_token import TokenRefresher
from libs.process_manifest_r3 import ManifestProcessor
from libs.traverse_manifest import ManifestTraversal
from libs.validation.validate_data_integrity import DataIntegrityValidator
from libs.validation.validate_referential_integrity import ManifestIntegrity
from libs.validation.validate_schema import SchemaValidator

logger = logging.getLogger()


class SingleManifestProcessor:

    def __init__(
        self,
        storage_url: str,
        payload_context: Context,
        data_integrity_validator: DataIntegrityValidator,
        referential_integrity_validator: ManifestIntegrity,
        manifest_processor: ManifestProcessor,
        schema_validator: SchemaValidator,
        token_refresher: TokenRefresher,
    ):
        """Init SingleManifestProcessor."""
        super().__init__()
        self.storage_url = storage_url
        self.payload_context = payload_context
        self.data_integrity_validator = data_integrity_validator
        self.referential_integrity_validator = referential_integrity_validator
        self.manifest_processor = manifest_processor
        self.schema_validator = schema_validator
        self.token_refresher = token_refresher

    def _process_record(self, entity_node: EntityNode) -> str:
        """
        Attempt to store a single record in Storage service.

        :param record: A record to be stored
        :return: Record id
        """
        try:
            logger.debug(f"Process entity {entity_node}")
            entity_node.replace_parents_surrogate_srns()
            record_id = self.manifest_processor.process_manifest_records(
                [entity_node.entity_info]
            )[FIRST_STORED_RECORD_INDEX]
            entity_node.system_srn = record_id
        except Exception as e:
            raise ProcessRecordError(entity_node.entity_info.entity, f"{e}"[:128])
        return record_id

    def _process_records(self, manifest_analyzer: ManifestAnalyzer) -> Tuple[List[str], List[dict]]:
        """
        Process each entity from entity queue created according to child-parent relationships
        between entities.
        Replace surrogate-keys of parents inside child entities with system-generated keys.

        :param manifest_analyzer: Object with proper queue of entities
        :return: List of saved ids and skipped ones.
        """
        record_ids = []
        skipped_ids = []
        for entity in manifest_analyzer.entity_queue():
            try:
                record_ids.append(self._process_record(entity))
            except ProcessRecordError as error:
                logger.warning(f"Can't process entity {entity}")
                manifest_analyzer.add_unprocessed_entity(entity)
                skipped_ids.append(error.skipped_entity)
        return record_ids, skipped_ids

    def process_manifest(self, manifest: dict, with_validation: bool) -> Tuple[
        List[str], List[dict]]:
        """Execute manifest validation then process it.

        Execution steps:
        1) validate referential manifest data integrity and remove invalid entities
        2) validate data integrity and prune to maintain dependency consistency
        3) initialize schema validator
        4) validate manifest file against common schema
        5) traverse manifest file and extract manifest entities
        6) validate extracted manifest entities
        7) create an ingestion queue according to child-parent relationships between entities
        8) process valid manifest entities one-by-one

        :param with_validation: Flag for validation steps
        :return: List of record ids and list of skipped entities.
        """
        skipped_ids = []
        if with_validation:
            manifest, not_valid_entities = \
                self.data_integrity_validator.validate_manifest_data_integrity(manifest)
            manifest_schema = self.schema_validator.validate_common_schema(manifest)
            skipped_ids.extend(not_valid_entities)
        else:
            manifest_schema = self.schema_validator.get_schema(
                self.schema_validator.get_manifest_kind(manifest))

        traversal = ManifestTraversal(manifest, manifest_schema)
        manifest_entities = traversal.traverse_manifest()

        # This set is used inside manifest_analyzer to know what entities must be before schema validation and integrity check.
        # Used to detect orphaned entities in manifest_analyzer
        original_manifest_entities_ids = set(
            entity.entity["id"] for entity in manifest_entities if entity.entity.get("id")
        )
        logger.debug(f"entities count: {len(manifest_entities)}")
        if with_validation:
            manifest_entities, not_valid_entities = \
                self.schema_validator.validate_manifest(manifest_entities)
            logger.debug(f"valid schema entities count: {len(manifest_entities)}")

            manifest, orphaned_entities = \
                self.referential_integrity_validator.ensure_integrity(manifest)
            logger.debug(f"valid referential integrity entities count: {len(manifest_entities)}")

            skipped_ids.extend(not_valid_entities)
            skipped_ids.extend(orphaned_entities)

        manifest_analyzer = ManifestAnalyzer(
            manifest_entities,
            self.token_refresher,
            original_manifest_entities_ids
        )

        record_ids, not_valid_ids = self._process_records(manifest_analyzer)
        skipped_ids.extend(not_valid_ids)

        logger.info(f"Processed ids {record_ids}")

        return record_ids, skipped_ids
