# Copyright © Amazon Web Services
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AWS Credential Module."""

import logging
from manifest_ingestion.providers.constants import AWS_CLOUD_PROVIDER
from manifest_ingestion.providers.factory import ProvidersFactory
from manifest_ingestion.providers.types import BaseCredentials
from manifest_ingestion.providers.aws.service_principal_util import get_service_principal_token
from tenacity import retry, stop_after_attempt
import os

logger = logging.getLogger(__name__)
RETRIES = 3

@ProvidersFactory.register(AWS_CLOUD_PROVIDER)
class AWSCredentials(BaseCredentials):
    """AWS Credential Provider"""

    def __init__(self):
        """Initialize AWS Credentials object"""
        self._access_token = None

    @retry(stop=stop_after_attempt(RETRIES))
    def refresh_token(self) -> str:
        """Refresh token.

        :return: Refreshed token
        :rtype: str
        """
        token = get_service_principal_token()
        self._access_token = token
        return self._access_token

    @property
    def access_token(self) -> str:
        """The access token.

        :return: Access token string.
        :rtype: str
        """
        return self._access_token

