#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import os
import logging
import sys

from functools import partial

import pytest
from file_paths import SURROGATE_MANIFEST_WELLBORE

sys.path.append(f"{os.getenv('AIRFLOW_SRC_DIR')}/plugins")
sys.path.append(f"{os.getenv('AIRFLOW_SRC_DIR')}/dags")

from libs.manifest_analyzer import ManifestAnalyzer, EntityNode
from libs.traverse_manifest import ManifestEntity
from libs.refresh_token import AirflowTokenRefresher
from libs.context import Context

logger = logging.getLogger()

TEST_FAKE_DATA = [
        {
            "id": "1",
            "parents": [],
        },
        {
            "id": "2",
            "parents": ["1"],
        },
        {
            "id": "3",
            "parents": ["1"],
        },
        {
            "id": "4",
            "parents": ["2"],
        },
        {
            "id": "5",
            "parents": ["1", "3"],
        },
        {
            "id": "7",
            "parents": ["1", "6"],
        },
        {
            "id": "9",
            "parents": ["7"]
        }
    ]

TEST_FAKE_DATA = [ManifestEntity(entity=i, schema="") for i in TEST_FAKE_DATA]

class TestManifestAnalyzer(object):

    @pytest.fixture()
    def manifest_analyzer(self):
        with open(SURROGATE_MANIFEST_WELLBORE) as f:
            data = json.load(f)
        context = Context(data_partition_id="test", app_key="test")
        token_refresher = AirflowTokenRefresher()
        return ManifestAnalyzer(data, "http://test.com", token_refresher, context)

    @pytest.fixture()
    def fake_data_manifest_analyzer(self, monkeypatch, data, is_in_storage: bool = False):
        context = Context(data_partition_id="test", app_key="test")
        token_refresher = AirflowTokenRefresher()
        monkeypatch.setattr(EntityNode, "get_parent_srns", self.mock_get_parent_srns)
        manifest_analyzer = ManifestAnalyzer(data, token_refresher)
        return manifest_analyzer

    def process_entity(self, entity: EntityNode) -> str:
        if "surrogate-key" in entity.srn:
            return f"system_srn: {entity.srn}"
        else:
            return entity.srn

    def index_in_queue_by_srn(
        self,
        manifest_analyzer: ManifestAnalyzer,
        queue: list,
        srn: str
    ):
        entity_node = manifest_analyzer.srn_node_table[srn]
        return queue.index(entity_node)

    @staticmethod
    def mock_get_parent_srns(obj: EntityNode, other_ids: set):
        parent_srns = set(obj.content.get("parents", []))
        return parent_srns

    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(TEST_FAKE_DATA, id="Fake data")
        ]
    )
    def test_queue_order(
        self,
        monkeypatch,
        fake_data_manifest_analyzer: ManifestAnalyzer,
        data: dict
    ):
        """
        Here we use array with simple objects where it's immediately seen who depends on whom.
        Check if queue return parents, then and only then their children.
        Check if there is no orphaned and their children in the queue (SRN 7 and SRN 9).
        """
        queue = list(fake_data_manifest_analyzer.entity_queue())
        index_in_queue = partial(self.index_in_queue_by_srn, fake_data_manifest_analyzer, queue)

        # check if child goes after all its parents in queue.
        assert index_in_queue("5") > index_in_queue("1") \
               and index_in_queue("5") > index_in_queue("3"), \
            "SRN 5 must follow parents: SRN 1 and 3"

        # check if orphans and their dependants are not in ingestion queue.
        for unprocessed_srn in ("7", "9"):
            unprocessed_entity = fake_data_manifest_analyzer.srn_node_table[unprocessed_srn]
            assert unprocessed_entity not in queue, \
                f"{unprocessed_entity} expected not to be in queue: {queue}"

    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(TEST_FAKE_DATA, id="Fake data")
        ]
    )
    def test_add_new_unporcessed(
        self,
        monkeypatch,
        fake_data_manifest_analyzer: ManifestAnalyzer,
        data: dict
    ):
        """
        Here we use array with simple objects where it's immediately seen who depends on whom.
        Imagine we can't process entity (e.g. Storage service can't save this entity).
        Then we must add this entity to unprocessed ones and traverse all the children of
        this entity marking them as unprocessed.
        They must disappear from the ingestion queue.
        """

        queue = fake_data_manifest_analyzer.entity_queue()
        unprocessed_entity = fake_data_manifest_analyzer.srn_node_table["3"]
        expected_unprocessed_entities = {"7", "9", "3", "5"}
        fake_data_manifest_analyzer.add_unprocessed_entity(unprocessed_entity)
        for entity in queue:
            assert entity not in expected_unprocessed_entities, \
                f"{entity} must be excluded from queue."

    def test_real_data(self):
        with open(SURROGATE_MANIFEST_WELLBORE) as f:
            data = json.load(f)
        data = [ManifestEntity(entity=e, schema="") for e in data]
        context = Context(data_partition_id="test", app_key="test")
        token_refresher = AirflowTokenRefresher()
        manifest_analyzer = ManifestAnalyzer(data, token_refresher)

        for entity in manifest_analyzer.entity_queue():
            entity.replace_parents_surrogate_srns()
            entity.system_srn = self.process_entity(entity)
            logger.info(f"Processed entity {json.dumps(entity.content, indent=2)}")

        logger.info("\n")
        logger.info(f"Processed entities: {manifest_analyzer.processed_entities}")
        logger.info(f"Unprocessed entities {manifest_analyzer.unprocessed_entities}")

    def test_real_data_with_orphaned(self):
        with open(SURROGATE_MANIFEST_WELLBORE) as f:
            data = json.load(f)
        data = [ManifestEntity(entity=e, schema="") for e in data]
        context = Context(data_partition_id="test", app_key="test")
        token_refresher = AirflowTokenRefresher()
        manifest_analyzer = ManifestAnalyzer(
            data,
            token_refresher,
            {"osdu:master-data--Wellbore:7587"}
        )

        for entity in manifest_analyzer.entity_queue():
            entity.replace_parents_surrogate_srns()
            entity.system_srn = self.process_entity(entity)
            logger.info(f"Processed entity {json.dumps(entity.content, indent=2)}")

        logger.info("\n")
        logger.info(f"Processed entities: {manifest_analyzer.processed_entities}")
        logger.info(f"Unprocessed entities {manifest_analyzer.unprocessed_entities}")
        assert "surrogate-key:wpc-1" in (e.srn for e in manifest_analyzer.unprocessed_entities)

    @pytest.mark.parametrize(
        "manifest,system_srn,expected_replaced_srns",
        [
            pytest.param(
                [
                    {
                        "id": "surrogate-key:wpc",
                        "ref": "surrogate-key:wpc2"
                    },
                    {
                        "id": "surrogate-key:wpc2"
                    }

                ],
                "test:work-product--WorkProduct:7b37ff0e13ac40a0ac35b7e5ca60e5a7",
                "test:work-product--WorkProduct:7b37ff0e13ac40a0ac35b7e5ca60e5a7:",
                id="Surrogate key"
            ),
            pytest.param(
                [
                    {
                        "id": "surrogate-key:wpc",
                        "ref": "osdu:reference-data--ResourceSecurityClassification:RESTRICTED:test:"
                    },
                    {
                        "id": "osdu:reference-data--ResourceSecurityClassification:RESTRICTED:test"
                    }

                ],
                "osdu:reference-data--ResourceSecurityClassification:RESTRICTED:test",
                "osdu:reference-data--ResourceSecurityClassification:RESTRICTED:test:",
                id="Real Ids"
            )
        ]
    )
    def test_update_parent_srn_with_system_srn(self, manifest, system_srn, expected_replaced_srns):
        data = [ManifestEntity(entity=e, schema="") for e in manifest]
        token_refresher = AirflowTokenRefresher()
        manifest_analyzer = ManifestAnalyzer(
            data,
            token_refresher
        )

        for entity in manifest_analyzer.entity_queue():
            entity.replace_parents_surrogate_srns()
            entity.system_srn = system_srn
            if entity.content.get("ref"):
                assert entity.content["ref"] == expected_replaced_srns
