#  Copyright 2020 Google LLC
#  Copyright 2020 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import json
import os
import sys

sys.path.append(f"{os.getenv('AIRFLOW_SRC_DIR')}/plugins")
sys.path.append(f"{os.getenv('AIRFLOW_SRC_DIR')}/dags")

from libs.manifest_analyzer import ManifestAnalyzer
from libs.processors.single_manifest_processor import SingleManifestProcessor
from libs.refresh_token import AirflowTokenRefresher
from libs.traverse_manifest import ManifestEntity
from libs.exceptions import EmptyManifestError
import pytest
import uuid
import requests

import mock_providers
from file_paths import SURROGATE_MANIFEST_WELLBORE, MANIFEST_GENERIC_PATH
from libs import process_manifest_r3

TENANT = "opendes"


class MockManifestProcessor:

    def process_manifest_records(self, records):
        return [i.entity.get("id", str(uuid.uuid4())) for i in records]


class MockDataIntegrityValidator:

    def validate_manifest_data_integrity(self, manifest):
        return manifest, []


class MockSchemaValidator:

    def get_schema(self, manifest):
        return {"properties": {"items": "a"}}

    def get_manifest_kind(self, manifest):
        return ""


class TestSingleManifestProcessor:

    @pytest.fixture
    def single_manifest_processor(self):
        token_refresher = AirflowTokenRefresher()
        return SingleManifestProcessor(
            storage_url="",
            manifest_processor=MockManifestProcessor(),
            token_refresher=token_refresher,
            data_integrity_validator="",
            schema_validator=MockSchemaValidator(),
            payload_context="",
            referential_integrity_validator=""
        )

    def test__process_valid_records(self, single_manifest_processor):
        token_refresher = AirflowTokenRefresher()
        with open(SURROGATE_MANIFEST_WELLBORE) as f:
            data = json.load(f)
        data = [ManifestEntity(entity=e, schema="") for e in data]
        manifest_analyzer = ManifestAnalyzer(data, token_refresher)
        records, skipped = single_manifest_processor._process_records(manifest_analyzer)
        assert records and not skipped

    def test__process_records_bad_response_from_storage(self, monkeypatch,
                                                        single_manifest_processor):
        def mock_bad_response(*args, **kwargs):
            raise Exception("Dummy exception")

        token_refresher = AirflowTokenRefresher()
        with open(SURROGATE_MANIFEST_WELLBORE) as f:
            data = json.load(f)
        data = [ManifestEntity(entity=e, schema="") for e in data]
        monkeypatch.setattr(MockManifestProcessor, "process_manifest_records", mock_bad_response)
        manifest_analyzer = ManifestAnalyzer(data, token_refresher)
        records, skipped = single_manifest_processor._process_records(manifest_analyzer)
        assert not records and skipped
